#include <assert.h>
#include <argp.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <sys/resource.h>
#include <bpf/libbpf.h>
#include <bpf/bpf.h>
#include <sys/select.h>
#include <unistd.h>
#include "paf.skel.h"
#include "pr.skel.h"
#include "procstat.skel.h"
#include "sysstat.skel.h"
#include "fraginfo.skel.h"
#include "numafraginfo.skel.h"
#include "memleak.skel.h"
#include "vmasnap.skel.h"
#include "drsnoop.skel.h"

#include "mem_watcher.h"
#include "fraginfo.h"
#include "oomkiller.skel.h"
#include "blazesym.h"

// 定义标志结构体
typedef struct
{
	int flag;
	const char *name;
} Flag;

// 定义所有组合修饰符和单独标志位
Flag gfp_combined_list[] = {
	{GFP_ATOMIC, "GFP_ATOMIC"},
	{GFP_KERNEL, "GFP_KERNEL"},
	{GFP_KERNEL_ACCOUNT, "GFP_KERNEL_ACCOUNT"},
	{GFP_NOWAIT, "GFP_NOWAIT"},
	{GFP_NOIO, "GFP_NOIO"},
	{GFP_NOFS, "GFP_NOFS"},
	{GFP_USER, "GFP_USER"},
	{GFP_DMA, "GFP_DMA"},
	{GFP_DMA32, "GFP_DMA32"},
	{GFP_HIGHUSER, "GFP_HIGHUSER"},
	{GFP_HIGHUSER_MOVABLE, "GFP_HIGHUSER_MOVABLE"},
	{GFP_TRANSHUGE_LIGHT, "GFP_TRANSHUGE_LIGHT"},
	{GFP_TRANSHUGE, "GFP_TRANSHUGE"},
};

Flag gfp_separate_list[] = {
	{___GFP_DMA, "___GFP_DMA"},
	{___GFP_HIGHMEM, "___GFP_HIGHMEM"},
	{___GFP_DMA32, "___GFP_DMA32"},
	{___GFP_MOVABLE, "___GFP_MOVABLE"},
	{___GFP_RECLAIMABLE, "___GFP_RECLAIMABLE"},
	{___GFP_HIGH, "___GFP_HIGH"},
	{___GFP_IO, "___GFP_IO"},
	{___GFP_FS, "___GFP_FS"},
	{___GFP_ZERO, "___GFP_ZERO"},
	{___GFP_ATOMIC, "___GFP_ATOMIC"},
	{___GFP_DIRECT_RECLAIM, "___GFP_DIRECT_RECLAIM"},
	{___GFP_KSWAPD_RECLAIM, "___GFP_KSWAPD_RECLAIM"},
	{___GFP_WRITE, "___GFP_WRITE"},
	{___GFP_NOWARN, "___GFP_NOWARN"},
	{___GFP_RETRY_MAYFAIL, "___GFP_RETRY_MAYFAIL"},
	{___GFP_NOFAIL, "___GFP_NOFAIL"},
	{___GFP_NORETRY, "___GFP_NORETRY"},
	{___GFP_MEMALLOC, "___GFP_MEMALLOC"},
	{___GFP_COMP, "___GFP_COMP"},
	{___GFP_NOMEMALLOC, "___GFP_NOMEMALLOC"},
	{___GFP_HARDWALL, "___GFP_HARDWALL"},
	{___GFP_THISNODE, "___GFP_THISNODE"},
	{___GFP_ACCOUNT, "___GFP_ACCOUNT"},
	{___GFP_ZEROTAGS, "___GFP_ZEROTAGS"},
	{___GFP_SKIP_KASAN_POISON, "___GFP_SKIP_KASAN_POISON"},
};

// static const int stack_map_max_entries = 10240; // 最大允许存储多少个stack_id

#define KERN_STACKID_FLAGS (0 | BPF_F_FAST_STACK_CMP)
#define USER_STACKID_FLAGS (0 | BPF_F_FAST_STACK_CMP | BPF_F_USER_STACK)


// static int attach_pid;
pid_t own_pid;


struct allocation
{
	int stack_id;
	__u64 size;
	size_t count;
};
// ============================= fraginfo====================================
struct order_entry
{
	struct order_zone okey;
	struct ctg_info oinfo;
};

int compare_entries(const void *a, const void *b)
{
	struct order_entry *entryA = (struct order_entry *)a;
	struct order_entry *entryB = (struct order_entry *)b;

	if (entryA->okey.zone_ptr != entryB->okey.zone_ptr)
	{
		return (entryA->okey.zone_ptr < entryB->okey.zone_ptr) ? -1 : 1;
	}
	else
	{
		return (entryA->okey.order < entryB->okey.order) ? -1 : 1;
	}
}

// ============================= fraginfo====================================

static volatile bool exiting = false;

#define __ATTACH_UPROBE(skel, sym_name, prog_name, is_retprobe)  \
	do                                                           \
	{                                                            \
		LIBBPF_OPTS(bpf_uprobe_opts, uprobe_opts,                \
					.func_name = #sym_name,                      \
					.retprobe = is_retprobe);                    \
		skel->links.prog_name = bpf_program__attach_uprobe_opts( \
			skel->progs.prog_name,                               \
			attach_pid,                                          \
			0,                                                   \
			&uprobe_opts);                                       \
	} while (false)

#define __CHECK_PROGRAM(skel, prog_name)                   \
	do                                                     \
	{                                                      \
		if (!skel->links.prog_name)                        \
		{                                                  \
			perror("no program attached for " #prog_name); \
			return -errno;                                 \
		}                                                  \
	} while (false)

#define __ATTACH_UPROBE_CHECKED(skel, sym_name, prog_name, is_retprobe) \
	do                                                                  \
	{                                                                   \
		__ATTACH_UPROBE(skel, sym_name, prog_name, is_retprobe);        \
		__CHECK_PROGRAM(skel, prog_name);                               \
	} while (false)

#define ATTACH_UPROBE(skel, sym_name, prog_name) __ATTACH_UPROBE(skel, sym_name, prog_name, false)
#define ATTACH_URETPROBE(skel, sym_name, prog_name) __ATTACH_UPROBE(skel, sym_name, prog_name, true)

#define ATTACH_UPROBE_CHECKED(skel, sym_name, prog_name) __ATTACH_UPROBE_CHECKED(skel, sym_name, prog_name, false)
#define ATTACH_URETPROBE_CHECKED(skel, sym_name, prog_name) __ATTACH_UPROBE_CHECKED(skel, sym_name, prog_name, true)

#define PROCESS_SKEL(skel, func)                                   \
	skel = func##_bpf__open();                                     \
	if (!skel)                                                     \
	{                                                              \
		fprintf(stderr, "Failed to open and load BPF skeleton\n"); \
		return 1;                                                  \
	}                                                              \
	process_##func(skel)

#define POLL_RING_BUFFER(rb, timeout, err)                  \
	while (!exiting)                                        \
	{                                                       \
		err = ring_buffer__poll(rb, timeout);               \
		if (err == -EINTR)                                  \
		{                                                   \
			err = 0;                                        \
			break;                                          \
		}                                                   \
		if (err < 0)                                        \
		{                                                   \
			printf("Error polling perf buffer: %d\n", err); \
			break;                                          \
		}                                                   \
	}

// 为 oomkiller 使用的宏，指定 map_name
#define LOAD_AND_ATTACH_SKELETON_WITH_MAP(skel, event, map_name)                                \
	do                                                                                         \
	{                                                                                          \
		err = event##_bpf__load(skel);                                                         \
		if (err)                                                                               \
		{                                                                                      \
			fprintf(stderr, "Failed to load and verify BPF skeleton\n");                       \
			goto event##_cleanup;                                                              \
		}                                                                                      \
                                                                                               \
		err = event##_bpf__attach(skel);                                                       \
		if (err)                                                                               \
		{                                                                                      \
			fprintf(stderr, "Failed to attach BPF skeleton\n");                                \
			goto event##_cleanup;                                                              \
		}                                                                                      \
                                                                                               \
		rb = ring_buffer__new(bpf_map__fd(skel->maps.map_name), handle_event_##event, NULL, NULL); \
		if (!rb)                                                                               \
		{                                                                                      \
			fprintf(stderr, "Failed to create ring buffer\n");                                 \
			goto event##_cleanup;                                                              \
		}                                                                                      \
	} while (0)

// 保留原有逻辑的宏
#define LOAD_AND_ATTACH_SKELETON(skel, event)                                                   \
	do                                                                                         \
	{                                                                                          \
		err = event##_bpf__load(skel);                                                         \
		if (err)                                                                               \
		{                                                                                      \
			fprintf(stderr, "Failed to load and verify BPF skeleton\n");                       \
			goto event##_cleanup;                                                              \
		}                                                                                      \
                                                                                               \
		err = event##_bpf__attach(skel);                                                       \
		if (err)                                                                               \
		{                                                                                      \
			fprintf(stderr, "Failed to attach BPF skeleton\n");                                \
			goto event##_cleanup;                                                              \
		}                                                                                      \
                                                                                               \
		rb = ring_buffer__new(bpf_map__fd(skel->maps.rb), handle_event_##event, NULL, NULL); \
		if (!rb)                                                                               \
		{                                                                                      \
			fprintf(stderr, "Failed to create ring buffer\n");                                 \
			goto event##_cleanup;                                                              \
		}                                                                                      \
	} while (0)

static struct env
{
    int time;            // 最大运行时间，单位为秒
    bool paf;            // 是否启用内存页面状态报告
    bool pr;             // 是否启用页面回收状态报告
    bool procstat;       // 是否启用进程内存状态报告
    bool sysstat;        // 是否启用系统内存状态报告
    bool memleak;        // 是否启用内核态/用户态内存泄漏检测
    bool fraginfo;       // 是否启用内存碎片信息
	bool numafraginfo;
    bool vmasnap;        // 是否启用虚拟内存区域信息
    bool drsnoop;
    bool kernel_trace;   // 是否启用内核态跟踪
    // bool print_time;     // 是否打印地址申请时间
    int interval;        // 打印间隔，单位为秒
    int duration;        // 运行时长，单位为秒
    bool part2;          // 是否启用系统内存状态报告的扩展部分
    bool oomkiller;      // 是否启用oomkiller事件处理

    long choose_pid;     // 选择的进程号
    bool rss;            // 是否打印进程页面信息
} env = {
    .time = 0,             // 0 表示无限运行
    .paf = false,          // 默认关闭内存页面状态报告
    .pr = false,           // 默认关闭页面回收状态报告
    .procstat = false,     // 默认关闭进程内存状态报告
    .sysstat = false,      // 默认关闭系统内存状态报告
    .memleak = false,      // 默认关闭内存泄漏检测
    .fraginfo = false,     // 默认关闭内存碎片信息
	.numafraginfo=false,
    .vmasnap = false,      // 默认关闭虚拟内存区域信息
    .drsnoop = false,
    .kernel_trace = true,  // 默认启用内核态跟踪
    // .print_time = false,   // 默认不打印地址申请时间
    .rss = false,          // 默认不打印进程页面信息
    .part2 = false,        // 默认关闭系统内存状态报告的扩展部分
    .oomkiller = false,    // 默认关闭oomkiller事件处理
    .choose_pid = 0,       // 默认不选择特定进程
    .interval = 1,         // 默认打印间隔为1秒
    .duration = 10,        // 默认持续运行10秒
};


const char argp_program_doc[] = "mem_watcher is in use ....\n";

static const struct argp_option opts[] = {
	{0, 0, 0, 0, "select function:", 1},

	{0, 0, 0, 0, "par:", 2},
	{"paf", 'a', 0, 0, "print paf (内存页面状态报告)"},


	{0, 0, 0, 0, "sysstat:", 6},
	{"sysstat", 's', 0, 0, "print sysstat (系统内存状态报告)"},


	{NULL, 'h', NULL, OPTION_HIDDEN, "show the full help"},
	{0},
};

static error_t parse_arg(int key, char *arg, struct argp_state *state)
{
	switch (key)
	{

	case 'a':
		env.paf = true;
		break;

	case 's':
		env.sysstat = true;
		break;

	default:
		return ARGP_ERR_UNKNOWN;
	}
	return 0;
}

static const struct argp argp = {
	.options = opts,
	.parser = parse_arg,
	.doc = argp_program_doc,
};


// Function prototypes

static void sig_handler(int sig);
static void setup_signals(void);


static int handle_event_paf(void *ctx, void *data, size_t data_sz);

static int handle_event_sysstat(void *ctx, void *data, size_t data_sz);

static void print_flag_modifiers(int flag);
static int process_paf(struct paf_bpf *skel_paf);

static int process_sysstat(struct sysstat_bpf *skel_sysstat);



// Main function
int main(int argc, char **argv)
{
    int err;
    struct paf_bpf *skel_paf;

    struct sysstat_bpf *skel_sysstat;


    err = argp_parse(&argp, argc, argv, 0, NULL, NULL);
    if (err)
        return err;

    own_pid = getpid();
    libbpf_set_strict_mode(LIBBPF_STRICT_ALL);

    setup_signals();

    if (env.paf)
    {
        PROCESS_SKEL(skel_paf, paf);
    }

    else if (env.sysstat)
    {
        PROCESS_SKEL(skel_sysstat, sysstat);
    }

    return 0;
}


int alloc_size_compare(const void *a, const void *b)
{
	const struct allocation *x = (struct allocation *)a;
	const struct allocation *y = (struct allocation *)b;

	// descending order

	if (x->size > y->size)
		return -1;

	if (x->size < y->size)
		return 1;

	return 0;
}


static void sig_handler(int sig)
{
	exiting = true;
	exit(EXIT_SUCCESS);
}

static void setup_signals(void)
{
	signal(SIGINT, sig_handler);
	signal(SIGTERM, sig_handler);
	signal(SIGALRM, sig_handler);
}

static void print_flag_modifiers(int flag)
{
	char combined[512] = {0}; // 用于保存组合修饰符
	char separate[512] = {0}; // 用于保存单独标志位

	// 检查组合修饰符
	for (int i = 0; i < sizeof(gfp_combined_list) / sizeof(gfp_combined_list[0]); ++i)
	{
		if ((flag & gfp_combined_list[i].flag) == gfp_combined_list[i].flag)
		{
			strcat(combined, gfp_combined_list[i].name);
			strcat(combined, " | ");
		}
	}

	// 移除最后一个 " | " 字符串的末尾
	if (strlen(combined) > 3)
	{
		combined[strlen(combined) - 3] = '\0';
	}

	// 检查单独标志位
	for (int i = 0; i < sizeof(gfp_separate_list) / sizeof(gfp_separate_list[0]); ++i)
	{
		if (flag & gfp_separate_list[i].flag)
		{
			strcat(separate, gfp_separate_list[i].name);
			strcat(separate, " | ");
		}
	}

	// 移除最后一个 " | " 字符串的末尾
	if (strlen(separate) > 3)
	{
		separate[strlen(separate) - 3] = '\0';
	}

	// 打印组合修饰符和单独标志位
	printf("%-50s %-100s\n", combined, separate);
}

static int handle_event_paf(void *ctx, void *data, size_t data_sz)
{
	const struct paf_event *e = data;
	struct tm *tm;
	char ts[32];
	time_t t;

	time(&t);
	tm = localtime(&t);
	strftime(ts, sizeof(ts), "%H:%M:%S", tm);

	printf("%-8lu %-8lu %-8lu %-8lu %-8x ",
		   e->min, e->low, e->high, e->present, e->flag);
	print_flag_modifiers(e->flag);
	printf("\n");

	return 0;
}



static int handle_event_sysstat(void *ctx, void *data, size_t data_sz)
{
	const struct sysstat_event *e = data;
	struct tm *tm;
	char ts[32];
	time_t t;

	time(&t);
	tm = localtime(&t);
	strftime(ts, sizeof(ts), "%H:%M:%S", tm);

	if (env.part2 == true)
		printf("%-8lu %-8lu %-8lu %-8lu %-8lu %-8lu %-8lu %-8lu %-8lu\n", e->slab_reclaimable + e->kernel_misc_reclaimable, e->slab_reclaimable + e->slab_unreclaimable, e->slab_reclaimable, e->slab_unreclaimable, e->unstable_nfs, e->writeback_temp, e->anon_thps, e->shmem_thps, e->pmdmapped);
	else
		printf("%-8lu %-8lu %-8lu %-8lu %-8lu %-8lu %-8lu %-8lu %-8lu %-8lu %-8lu %-8lu\n", e->anon_active + e->file_active, e->file_inactive + e->anon_inactive, e->anon_active, e->anon_inactive, e->file_active, e->file_inactive, e->unevictable, e->file_dirty, e->writeback, e->anon_mapped, e->file_mapped, e->shmem);

	return 0;
}


// Functions to process different BPF programs
static int process_paf(struct paf_bpf *skel_paf)
{
	int err;
	struct ring_buffer *rb;

	LOAD_AND_ATTACH_SKELETON(skel_paf, paf);

	printf("%-8s %-8s  %-8s %-8s %-8s\n", "MIN", "LOW", "HIGH", "PRESENT", "FLAG");

	POLL_RING_BUFFER(rb, 1000, err);

paf_cleanup:
	ring_buffer__free(rb);
	paf_bpf__destroy(skel_paf);
	return err;
}



static int process_sysstat(struct sysstat_bpf *skel_sysstat)
{
	int err;
	struct ring_buffer *rb;

	LOAD_AND_ATTACH_SKELETON(skel_sysstat, sysstat);

	if (env.part2)
	{
		printf("%-8s %-8s %-8s %-8s %-8s %-8s %-8s %-8s %-8s\n", "KRECLM", "SLAB", "SRECLM", "SUNRECL", "NFSUNSTB", "WRITEBACKTMP", "KMAP", "UNMAP", "PAGE");
	}
	else
	{
		printf("%-8s %-8s %-8s %-8s %-8s %-8s %-8s %-8s %-8s %-8s %-8s %-8s\n", "ACTIVE", "INACTVE", "ANON_ACT", "ANON_INA", "FILE_ACT", "FILE_INA", "UNEVICT", "DIRTY", "WRITEBK", "ANONPAG", "MAP", "SHMEM");
	}
	POLL_RING_BUFFER(rb, 1000, err);

sysstat_cleanup:
	ring_buffer__free(rb);
	sysstat_bpf__destroy(skel_sysstat);
	return err;
}


// ================================================== fraginfo====================================================================
// compute order
static int __fragmentation_index(unsigned int order, long unsigned int total, long unsigned int suitable, long unsigned int free)
{
	unsigned long requested = 1UL << order;
	// 无可用内存返回0
	if (order > MAX_ORDER)
		return 0;
	if (!total)
		return 0;
	// 有可用内存返回-1000
	if (suitable)
		return -1000;
	double res1, res2;
	res1 = (double)(free * 1000ULL) / requested;
	// res1 +=1000;
	res2 = (double)res1 / total;
	return res2;
}
static int unusable_free_index(unsigned int order, long unsigned int total, long unsigned int suitable, long unsigned int free)
{
	/* No free memory is interpreted as all free memory is unusable */
	if (free == 0)
		return 1000;

	/*
	 * Index should be a value between 0 and 1. Return a value to 3
	 * decimal places.
	 *
	 * 0 => no fragmentation
	 * 1 => high fragmentation
	 */
	long unsigned int res1 = free - (suitable << order);
	double res = (res1 * 1000ULL) / free;
	return res;
}
void print_zones(int fd)
{
	struct zone_info zinfo;
	__u64 key = 0, next_key;
	printf("%-20s %-20s %-25s %-20s %-20s", " COMM", "ZONE_PTR", "ZONE_PFN ", " SUM_PAGES", "FACT_PAGES ");
	printf("\n");
	while (bpf_map_get_next_key(fd, &key, &next_key) == 0)
	{
		bpf_map_lookup_elem(fd, &next_key, &zinfo);
		printf(" %-15s 0x%-25llx %-25llu %-20llu %-15llu\n", zinfo.comm, zinfo.zone_ptr, zinfo.zone_start_pfn, zinfo.spanned_pages, zinfo.present_pages);
		key = next_key;
	}
}
void print_nodes(int fd)
{
	struct pgdat_info pinfo;
	__u64 key = 0, next_key;
	printf(" Node ID          PGDAT_PTR       NR_ZONES \n");
	while (bpf_map_get_next_key(fd, &key, &next_key) == 0)
	{
		bpf_map_lookup_elem(fd, &next_key, &pinfo);
		printf(" %5d       0x%llx  %5d\n",
			   pinfo.node_id, pinfo.pgdat_ptr, pinfo.nr_zones);
		key = next_key;
	}
}
void print_orders(int fd)
{
	struct order_zone okey = {};
	struct ctg_info oinfo;
	struct order_entry entries[256];
	int entry_count = 0;

	while (bpf_map_get_next_key(fd, &okey, &okey) == 0)
	{
		if (bpf_map_lookup_elem(fd, &okey, &oinfo) == 0)
		{
			entries[entry_count].okey = okey;
			entries[entry_count].oinfo = oinfo;
			entry_count++;
		}
	}

	// 排序
	qsort(entries, entry_count, sizeof(struct order_entry), compare_entries);

	// 打印排序后的
	printf(" Order     Zone_PTR                Free Pages         Free Blocks Total    Free Blocks Suitable      SCOREA     SCOREB\n");
	for (int i = 0; i < entry_count; i++)
	{
		int res = __fragmentation_index(entries[i].okey.order, entries[i].oinfo.free_blocks_total, entries[i].oinfo.free_blocks_suitable, entries[i].oinfo.free_pages);
		int tmp = unusable_free_index(entries[i].okey.order, entries[i].oinfo.free_blocks_total, entries[i].oinfo.free_blocks_suitable, entries[i].oinfo.free_pages);
		// int part1 = res / 1000;
		// int dec1 = res % 1000;
		int part2 = tmp / 1000;
		int dec2 = tmp % 1000;
		printf(" %-8u 0x%-25llx %-20lu %-20lu %-20lu %d   %d.%03d\n",
			   entries[i].okey.order, entries[i].okey.zone_ptr, entries[i].oinfo.free_pages,
			   entries[i].oinfo.free_blocks_total, entries[i].oinfo.free_blocks_suitable, res, part2, dec2);
	}
}

