// SPDX-License-Identifier: (LGPL-2.1 OR BSD-2-Clause)
/* Copyright (c) 2020 Facebook */
#include <argp.h>
#include <signal.h>
#include <stdio.h>
#include <time.h>
#include <sys/resource.h>
#include <bpf/libbpf.h>
#include <unistd.h>
#include "write.skel.h"
#include "write.h"

pid_t own_pid;

static int libbpf_print_fn(enum libbpf_print_level level, const char *format, va_list args)
{
	return vfprintf(stderr, format, args);
}

static volatile bool exiting;

static void sig_handler(int sig)
{
	exiting = true;
}

static int write_event(void *ctx, void *data, size_t data_sz)
{
   const struct sys_call_info *sc_if = data;
    struct tm *tm;
    char ts[32];
    time_t t;
    time(&t);
    tm = localtime(&t);
    strftime(ts, sizeof(ts), "%H:%M:%S", tm);
	printf("ts:%-8s  pid:%-7d  inode_number:%-7ld   cout:%-7ld   real_count:%-7ld\n", ts, sc_if->pid,sc_if->inode_number,sc_if->count,sc_if->real_count);
    return 0;
}
int main(int argc, char **argv)
{

	struct ring_buffer *rb = NULL;
	struct write_bpf *skel;
	int err;

	// int input_ic_sys_call_nr;
	own_pid = getpid();

	libbpf_set_strict_mode(LIBBPF_STRICT_ALL);
	/* Set up libbpf errors and debug info callback */
	libbpf_set_print(libbpf_print_fn);

	/* Cleaner handling of Ctrl-C */
	signal(SIGINT, sig_handler);
	signal(SIGTERM, sig_handler);

	/* Load and verify BPF application */
	skel = write_bpf__open();
	if (!skel)
	{
		fprintf(stderr, "Failed to open and load BPF skeleton\n");
		return 1;
	}
	/* -----------------------------这中间可以实现用户态c程序送入值进入ebpf程序，代码块位置不能变动------- */
	// printf("请输入你要拦截的系统调用的调用号码(范围0-517): ");
	// scanf("%d", &input_ic_sys_call_nr);
	// skel->bss->ic_sys_call_nr = input_ic_sys_call_nr;
	// skel->bss->own_pid = own_pid;

	/* ------------------------------------------------------------------------------------------------- */

	/* Load & verify BPF programs */
	err = write_bpf__load(skel);
	if (err)
	{
		fprintf(stderr, "Failed to load and verify BPF skeleton\n");
		goto cleanup;
	}

	/* Attach tracepoints */
	err = write_bpf__attach(skel);
	if (err)
	{
		fprintf(stderr, "Failed to attach BPF skeleton\n");
		goto cleanup;
	}

	/* Set up ring buffer polling */
	rb = ring_buffer__new(bpf_map__fd(skel->maps.rb), write_event, NULL, NULL);
	if (!rb)
	{
		err = -1;
		fprintf(stderr, "Failed to create ring buffer\n");
		goto cleanup;
	}

	while (!exiting)
	{
		err = ring_buffer__poll(rb, 100 /* timeout, ms */);
		/* Ctrl-C will cause -EINTR */
		if (err == -EINTR)
		{
			err = 0;
			break;
		}
		if (err < 0)
		{
			printf("Error polling perf buffer: %d\n", err);
			break;
		}
	}

cleanup:
	/* Clean up */
	ring_buffer__free(rb);
	write_bpf__destroy(skel);
	return err < 0 ? -err : 0;
}
